
# check for cda exception
# this is not correct. Exception can be raised on python's side, not in C
cdef inline int cda_check_exception(int code) except -1:
    if code < 0:
        raise Exception("cda error: %s, errcode: %s" % (cda_last_err(), code))
    return 0

# C callback function for ref's (channels)
cdef void evproc_rslvstat(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                          void *info_ptr, void *privptr2) with gil:
    cdef BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    chan.rslv_stat = <long>info_ptr # resolve status passed through this pointer

    if chan.rslv_stat == CDA_RSLVSTAT_NOTFOUND:
        chan.rslv_str = 'not found'
    elif chan.rslv_stat == CDA_RSLVSTAT_SEARCHING:
        chan.rslv_str = 'searching'
    elif chan.rslv_stat == CDA_RSLVSTAT_FOUND:
        chan.rslv_str = 'found'
    chan.resolve.emit(chan)


cdef void evproc_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        cx_time_t timestr
        rflags_t rflags
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
    cda_check_exception( cda_get_ref_stat(ref, &rflags, &timestr) )

    chan.prev_time = chan.time
    chan.time = <int64>timestr.sec * 1000000 + timestr.nsec / 1000
    chan.cb()


cdef void quant_update(int uniq, void *privptr1, cda_dataref_t ref, int reason,
                        void *info_ptr, void *privptr2) with gil:
    cdef:
        int res
        CxAnyVal_t quant_raw
        cxdtype_t quant_dtype
        BaseChan chan = <BaseChan>(<event*>privptr2).objptr
        double quant_draw, val, shift

    res = cda_quant_of_ref(chan.ref, &quant_raw, &quant_dtype)
    quant_draw = aval_value(&quant_raw, quant_dtype)
    if quant_draw == 0:
        chan.quant = 0
        return
    else:
        res = cda_rd_convert(ref, 0, &shift)
        res = cda_rd_convert(ref, quant_draw, &val)
        chan.quant = val - shift


# wrapper-class for low-level functions and channel registration
cdef class BaseChan(CdaObject):
    cdef readonly:
        cda_dataref_t ref
        str name
        int max_nelems
        cxdtype_t dtype
        size_t itemsize
        int64 time, prev_time
        int found # -1 - not found, 0 - unknown, 1 found
        long rslv_stat
        str rslv_str
        double quant
        int registered, first_cycle
        Context context

    IF SIGNAL_IMPL=='sl':
        cdef readonly:
            Signal valueMeasured, valueChanged, resolve
    ELIF SIGNAL_IMPL=='Qt':
        cdef:
            object c_valueChanged, c_valueMeasured, c_resolve
            public object valueChanged, valueMeasured, resolve

    def __init__(self, str name, cxdtype_t dtype=CXDTYPE_DOUBLE, int max_nelems=1, **kwargs):
        CdaObject.__init__(self)
        self.context = kwargs.get('context', default_context)

        IF SIGNAL_IMPL=='sl':
            self.valueChanged, self.valueMeasured, self.resolve = Signal(), Signal(), Signal()
        ELIF SIGNAL_IMPL=='Qt':
            self.c_valueChanged, self.c_valueMeasured, self.c_resolve = SignalContainer(), SignalContainer(), SignalContainer()
            self.valueChanged, self.valueMeasured, self.resolve = self.c_valueChanged.signal, self.c_valueMeasured.signal, self.c_resolve.signal

        b_name = name.encode("ascii")
        cdef:
            char *c_name = b_name
            int ret
            unsigned int options = 0

        # may be cycle?
        if kwargs.get('private', False):
            options += CDA_DATAREF_OPT_PRIVATE

        if kwargs.get('no_rd_conv', False):
            options += CDA_DATAREF_OPT_NO_RD_CONV

        if kwargs.get('shy', False):
            options += CDA_DATAREF_OPT_SHY

        if kwargs.get('find_only', False):
            options += CDA_DATAREF_OPT_FIND_ONLY

        if kwargs.get('on_update', False):
            options += CDA_DATAREF_OPT_ON_UPDATE

        if kwargs.get('no_wr_wait', False):
            options += CDA_DATAREF_OPT_NO_WR_WAIT

        if kwargs.get('debug', False):
            options += CDA_DATAREF_OPT_DEBUG

        ret = cda_add_chan((<Context>self.context).cid, NULL, c_name, options, dtype, max_nelems,
                           0, <cda_dataref_evproc_t>NULL, NULL)
        cda_check_exception(ret)
        self.ref, self.name, self.dtype, self.max_nelems, self.itemsize =\
            ret, name, dtype, max_nelems, cx.sizeof_cxdtype(dtype)

        (<Context>self.context).save_chan(<void*>self)

        # events which can be regular
        self.add_event(CDA_REF_EVMASK_RSLVSTAT, <void*>evproc_rslvstat, <void*>self, NULL)
        self.add_event(CDA_REF_EVMASK_QUANTCHG, <void*>quant_update, <void*>self, NULL)
        self.add_event(CDA_REF_EVMASK_UPDATE, <void*>evproc_update, <void*>self, NULL)

        # initialization events (will be unregistered when get)

        self.registered, self.first_cycle = True, True
        self.rslv_stat = CDA_RSLVSTAT_SEARCHING
        self.rslv_str = 'searching'

    def __dealloc__(self):
        if self.registered:
            (<Context>self.context).drop_chan(<void*>self)
            cda_del_chan(self.ref)
            self.registered = 0

    def __str__(self):
        return '<cda_chan: ref=%d, name=%s>' % (self.ref, self.name)

    def short_name(self):
        a = self.name.split('.')
        return a[len(a)-1].split('@')[0]

    cdef void cb(self):
        #empty callback for overrideing
        pass

    cdef int snd_data(self, cxdtype_t dtype, int nelems, void* data_p):
        cdef int res = cda_snd_ref_data(self.ref, dtype, nelems, data_p)
        cda_check_exception(res)
        return res

    cdef int get_data(self, size_t ofs, size_t size, void* buf):
        cdef int res = cda_get_ref_data(self.ref, ofs, size, buf)
        cda_check_exception(res)
        return res

    cdef int current_nelems(self):
        return cda_current_nelems_of_ref(self.ref)

    cdef void get_src(self, const char **src_p):
        cda_check_exception( cda_src_of_ref(self.ref, src_p) )

    cdef void register_event(self, event *ev):
        cda_check_exception( cda_add_dataref_evproc(self.ref, ev.evmask, <cda_dataref_evproc_t>ev.evproc, ev) )

    cdef void unregister_event(self, event *ev):
        cda_check_exception( cda_del_dataref_evproc(self.ref, ev.evmask, <cda_dataref_evproc_t>ev.evproc, ev) )







        # CXRF_SERVER_MASK
        # CXRF_CLIENT_MASK
        #
        # #/* Server-side flags -- CXRF_* */
        # CXRF_CAMAC_NO_X       #// No 'X' from device
        # CXRF_CAMAC_NO_Q       #// No 'Q' from device
        #
        # CXRF_IO_TIMEOUT       #// I/O timeout expired
        # CXRF_REM_C_PROBL      #// Remote controller problem
        #
        # CXRF_OVERLOAD         #// Input channel overload
        #
        # CXRF_UNSUPPORTED      #// Unsupported feature/channel
        #
        # CXRF_INVAL            #// Invalid parameter
        # CXRF_WRONG_DEV        #// Wrong device
        # CXRF_CFG_PROBL        #// Configuration problem
        # CXRF_DRV_PROBL        #// Driver internal problem
        # CXRF_NO_DRV           #// Driver loading problem
        # CXRF_OFFLINE          #// Device is offline
        #
        # CXRF_SERVER_HWERR_MASK
        # CXRF_SERVER_SFERR_MASK
        #
        # #/* Client-side flags -- CXCF_* */
        # #/* cda -- upwards from 16 */
        # CXCF_FLAG_CALCERR     #// Formula calculation error
        # CXCF_FLAG_DEFUNCT     #// Defunct channel
        # CXCF_FLAG_OTHEROP     #// Other operator is active
        # CXCF_FLAG_PRGLYCHG    #// Channel was programmatically changed
        # CXCF_FLAG_NOTFOUND    #// Channel not found
        # CXCF_FLAG_COLOR_WEIRD #// Value is weird
        #
        # CXCF_FLAG_CDA_MASK
        #
        # #/* Cdr -- downwards from 31 */
        # CXCF_FLAG_ALARM_ALARM  #// Alarm!
        # CXCF_FLAG_ALARM_RELAX  #// Relaxing after alarm
        # CXCF_FLAG_COLOR_RED    #// Value in red zone
        # CXCF_FLAG_COLOR_YELLOW #// Value in yellow zone
        #
        # CXCF_FLAG_CDR_MASK
        #
        # #/* General classification */
        # CXCF_FLAG_HWERR_MASK
        # CXCF_FLAG_SFERR_MASK
        # CXCF_FLAG_SYSERR_MASK
        #
        # CXCF_FLAG_ALARM_MASK
        # CXCF_FLAG_COLOR_MASK
        # CXCF_FLAG_COLOR_WEIRD
        # CXCF_FLAG_COLOR_RED
        # CXCF_FLAG_COLOR_YELLOW
        #
        # CXCF_FLAG_4WRONLY_MASK


cdef list text_flags(cx.rflags_t rflags):
    ret = []
    if (rflags & cx.CXRF_CAMAC_NO_X) > 0:
        ret.append('No X from CAMAC device')

    if (rflags & cx.CXRF_CAMAC_NO_Q) > 0:
        ret.append('No Q from CAMAC device')

    if (rflags & cx.CXRF_IO_TIMEOUT) > 0:
        ret.append('I/O timeout expired')

    if (rflags & cx.CXRF_REM_C_PROBL) > 0:
        ret.append('Remote controller problem')

    if (rflags & cx.CXRF_OVERLOAD) > 0:
        ret.append('Input channel overload')

    if (rflags & cx.CXRF_UNSUPPORTED) > 0:
        ret.append('Unsupported feature/channel')

    if (rflags & cx.CXRF_INVAL) > 0:
        ret.append('Invalid parameter')

    if (rflags & cx.CXRF_WRONG_DEV) > 0:
        ret.append('Wrong device')

    if (rflags & cx.CXRF_CFG_PROBL) > 0:
        ret.append('Configuration problem')

    if (rflags & cx.CXRF_DRV_PROBL) > 0:
        ret.append('Driver internal problem')

    if (rflags & cx.CXRF_NO_DRV) > 0:
        ret.append('Driver loading problem')

    if (rflags & cx.CXRF_OFFLINE) > 0:
        ret.append('Device is offline')

    if (rflags & cx.CXCF_FLAG_CALCERR) > 0:
        ret.append('Formula calculation error')

    if (rflags & cx.CXCF_FLAG_DEFUNCT) > 0:
        ret.append('Defunct channel')

    if (rflags & cx.CXCF_FLAG_OTHEROP) > 0:
        ret.append('Other operator is active')

    if (rflags & cx.CXCF_FLAG_PRGLYCHG) > 0:
        ret.append('Channel was programmatically changed')

    if (rflags & cx.CXCF_FLAG_NOTFOUND) > 0:
        ret.append('Channel not found')

    if (rflags & cx.CXCF_FLAG_COLOR_WEIRD) > 0:
        ret.append('Value is weird')

    if (rflags & cx.CXCF_FLAG_ALARM_ALARM) > 0:
        ret.append('Alarm!')

    if (rflags & cx.CXCF_FLAG_ALARM_RELAX) > 0:
        ret.append('Relaxing after alarm')

    if (rflags & cx.CXCF_FLAG_COLOR_RED) > 0:
        ret.append('Value in red zone')

    if (rflags & cx.CXCF_FLAG_COLOR_YELLOW) > 0:
        ret.append('Value in yellow zone')

    return ret


# data types declarations and utility functions

CXDTYPE_UNKNOWN = cx.CXDTYPE_UNKNOWN
CXDTYPE_INT8 = cx.CXDTYPE_INT8
CXDTYPE_INT16 = cx.CXDTYPE_INT16
CXDTYPE_INT32 = cx.CXDTYPE_INT32
CXDTYPE_INT64 = cx.CXDTYPE_INT64
CXDTYPE_UINT8 = cx.CXDTYPE_UINT8
CXDTYPE_UINT16 = cx.CXDTYPE_UINT16
CXDTYPE_UINT32 = cx.CXDTYPE_UINT32
CXDTYPE_UINT64 = cx.CXDTYPE_UINT64
CXDTYPE_SINGLE = cx.CXDTYPE_SINGLE
CXDTYPE_DOUBLE = cx.CXDTYPE_DOUBLE
CXDTYPE_TEXT = cx.CXDTYPE_TEXT
CXDTYPE_UCTEXT = cx.CXDTYPE_UCTEXT

cx_dtype_map = {
    'int8': CXDTYPE_INT8,
    'int16': CXDTYPE_INT16,
    'int32': CXDTYPE_INT32,
    'int': CXDTYPE_INT32,
    'int64': CXDTYPE_INT64,
    'uint8': CXDTYPE_UINT8,
    'uint16': CXDTYPE_UINT16,
    'uint32': CXDTYPE_UINT32,
    'uint': CXDTYPE_UINT32,
    'uint64': CXDTYPE_UINT64,
    'single': CXDTYPE_SINGLE,
    'float': CXDTYPE_SINGLE,
    'double': CXDTYPE_DOUBLE,
    'text': CXDTYPE_TEXT,
    'str': CXDTYPE_TEXT,
    'uctext': CXDTYPE_UCTEXT,
    #'unknown': CXDTYPE_UNKNOWN,
}

# function for pythonize cx-any-val
# not to effective, but not a bottleneck
cdef:
    aval_value(CxAnyVal_t *aval, cxdtype_t dtype):
        if dtype == cx.CXDTYPE_DOUBLE: return aval.f64
        if dtype == cx.CXDTYPE_INT32:  return aval.i32
        if dtype == cx.CXDTYPE_INT8:   return aval.i8
        if dtype == cx.CXDTYPE_INT16:  return aval.i16
        if dtype == cx.CXDTYPE_INT64:  return aval.i64
        if dtype == cx.CXDTYPE_UINT8:  return aval.u8
        if dtype == cx.CXDTYPE_UINT16: return aval.u16
        if dtype == cx.CXDTYPE_UINT32: return aval.u32
        if dtype == cx.CXDTYPE_UINT64: return aval.u64
        if dtype == cx.CXDTYPE_SINGLE: return aval.f32
        if dtype == cx.CXDTYPE_TEXT:   return aval.c8
        if dtype == cx.CXDTYPE_UCTEXT: return aval.c32
        return None

    void aval_setvalue(value, CxAnyVal_t *aval, cxdtype_t dtype):
        if dtype == cx.CXDTYPE_DOUBLE: aval.f64 = value
        if dtype == cx.CXDTYPE_INT32:  aval.i32 = value
        if dtype == cx.CXDTYPE_INT8:   aval.i8 = value
        if dtype == cx.CXDTYPE_INT16:  aval.i16 = value
        if dtype == cx.CXDTYPE_INT64:  aval.i64 = value
        if dtype == cx.CXDTYPE_UINT8:  aval.u8 = value
        if dtype == cx.CXDTYPE_UINT16: aval.u16 = value
        if dtype == cx.CXDTYPE_UINT32: aval.u32 = value
        if dtype == cx.CXDTYPE_UINT64: aval.u64 = value
        if dtype == cx.CXDTYPE_SINGLE: aval.f32 = value
        if dtype == cx.CXDTYPE_TEXT:   aval.c8 = value
        if dtype == cx.CXDTYPE_UCTEXT: aval.c32 = value

    # dtype conversion numpy <--> CX
    # there are no text or unicode text.
    # Cause text and unicode text can't be mixed with numbers in a good way
    object cxdtype2np(cxdtype_t dtype):
        if dtype == cx.CXDTYPE_DOUBLE: return np.double
        if dtype == cx.CXDTYPE_INT32:  return np.int32
        if dtype == cx.CXDTYPE_INT8:   return np.int8
        if dtype == cx.CXDTYPE_INT16:  return np.int16
        if dtype == cx.CXDTYPE_INT64:  return np.int64
        if dtype == cx.CXDTYPE_UINT8:  return np.uint8
        if dtype == cx.CXDTYPE_UINT16: return np.uint16
        if dtype == cx.CXDTYPE_UINT32: return np.uint32
        if dtype == cx.CXDTYPE_UINT64: return np.uint64
        if dtype == cx.CXDTYPE_SINGLE: return np.single
        return None

    cxdtype_t np2cxdtype(object dtype):
        if dtype == np.double:      return cx.CXDTYPE_DOUBLE
        if dtype == np.int32:       return cx.CXDTYPE_INT32
        if dtype == np.int8:        return cx.CXDTYPE_INT8
        if dtype == np.int16:       return cx.CXDTYPE_INT16
        if dtype == np.int64:       return cx.CXDTYPE_INT64
        if dtype == np.uint8:       return cx.CXDTYPE_UINT8
        if dtype == np.uint16:      return cx.CXDTYPE_UINT16
        if dtype == np.uint32:      return cx.CXDTYPE_UINT32
        if dtype == np.uint64:      return cx.CXDTYPE_UINT64
        if dtype == np.single:      return cx.CXDTYPE_SINGLE
        return 0




from cx4.cxscheduler cimport *

cpdef main_loop():
    sl_main_loop()

cpdef break_():
    sl_break()


#  required to react to signals in case of daemon
def select_interrupt_action():
    pass

cdef void at_select():
    select_interrupt_action()

sl_set_select_behaviour(<sl_at_select_proc>NULL, <sl_at_select_proc>at_select, 0)


import numpy as np
cimport numpy as np
from cpython.version cimport PY_MAJOR_VERSION
from libc.stdlib cimport realloc, free, malloc
from libc.string cimport memmove
from cx4 cimport cx
from cx4.cda cimport *

RSLVSTAT_NOTFOUND = CDA_RSLVSTAT_NOTFOUND
RSLVSTAT_SEARCHING = CDA_RSLVSTAT_SEARCHING
RSLVSTAT_FOUND = CDA_RSLVSTAT_FOUND


np.import_array()

# vector-data channel class
cdef class VChan(BaseChan):
    cdef:
        readonly np.ndarray val, buf_val
        readonly object npdtype

    def __init__(self, str name, cxdtype_t dtype=CXDTYPE_DOUBLE, int max_nelems=1, **kwargs):
        BaseChan.__init__(self, name, dtype, max_nelems, **kwargs)
        self.npdtype = cxdtype2np(dtype)
        self.buf_val = np.zeros(max_nelems, self.npdtype, order='C')
        self.val = np.zeros(0, self.npdtype, order='C')

    cdef void cb(self):
        c_len = self.get_data(0, self.itemsize * self.max_nelems, <void*>self.buf_val.data)
        nelems_read = int(c_len/self.itemsize)
        self.val = self.buf_val[:nelems_read]
        self.valueMeasured.emit(self)
        self.valueChanged.emit(self)

    cpdef void setValue(self, np.ndarray value):
        if value.size > self.max_nelems:
            raise Exception('value size greater than channel.max_nelems')
        dtype = np2cxdtype(value.dtype)
        self.snd_data(dtype, value.size, <void*>value.data)



__all__ = ['Context', 'StrChan', 'DChan', 'IChan', 'Chan', 'VChan',
           'CXDTYPE_UNKNOWN', 'CXDTYPE_INT8', 'CXDTYPE_INT16', 'CXDTYPE_INT32', 'CXDTYPE_INT64', 'CXDTYPE_UINT8','CXDTYPE_UINT16',
           'CXDTYPE_UINT32', 'CXDTYPE_UINT64', 'CXDTYPE_SINGLE', 'CXDTYPE_DOUBLE', 'CXDTYPE_TEXT', 'CXDTYPE_UCTEXT',
           'RSLVSTAT_NOTFOUND', 'RSLVSTAT_SEARCHING', 'RSLVSTAT_FOUND',
           'ChanFactory', 'cfactory',
    ]


include 'imports.pxi'

include 'scheduler.pxi'
include 'signal.pxi'
include 'timer.pxi'

# conpile-time define for contitional compilation
DEF SIGNAL_IMPL='sl'

include 'cda_common.pxi'

include 'gw.pxi'
__all__ += ['PassGW']

# add cx scheduler specific simbols
__all__ += ['main_loop', 'break_', 'Timer', 'Signal']


# Signal container to bypass inheritance from QObject
# to be encapsulated in Qt channel-like classes

class SignalContainer(QObject):
    signal = pyqtSignal(object)

    def __init__(self):
        super(SignalContainer, self).__init__()



#!/usr/bin/env python3

from pycx4.version import __version__
import subprocess

subprocess.run(['python3', 'setup.py', 'sdist'])
subprocess.run(['twine', 'upload', 'dist/pycx4-' + __version__ + '.tar.gz'])



# PyQtX reference is only here, may be nedd to do it more universal

from PyQt4.QtCore import QObject, pyqtSignal
from PyQt4.QtNetwork import QAbstractSocket

include 'imports.pxi'


include 'qt_signalcontainer.pxi'

# conpile-time define for contitional compilation
DEF SIGNAL_IMPL='Qt'

include 'cda_common.pxi'


from cx4.cxscheduler cimport sl_tout_proc, sl_enq_tout_after, sl_tid_t, sl_deq_tout

cdef void sltimer_proc(int uniq, void *privptr1, sl_tid_t tid, void *privptr2) with gil:
    cdef Timer t = <Timer>privptr2
    if t.repeat:
        t.tid = sl_enq_tout_after(0, NULL, t.usecs, sltimer_proc, privptr2)
    else:
        t.active = 0
    t.timeout.emit(t)



cdef class Timer:
    cdef readonly:
        int usec
        int repeat
        int active
        sl_tid_t tid
        Signal timeout

    def __init__(self, int msec=1000):
        self.setInterval(msec)
        self.repeat = 1
        self.timeout = Signal()

    cpdef stop(self):
        if self.active == 1:
            self.active = 0
            sl_deq_tout(self.tid)

    cpdef start(self):
        if self.\
                active == 1:
            return
        self.active = 1
        self.tid = sl_enq_tout_after(0, NULL, self.usecs, sltimer_proc, <void*>self)

    cpdef singleShot(self, int msec=0, proc=None):
        if msec > 0:
            self.setInterval(msec)
        if proc is not None:
            self.timeout.connect(proc)
        self.repeat = 0
        self.start()

    cpdef int interval(self):
        return <int>(self.usec/1000)

    cpdef void setInterval(self, int msec):
        self.usec = 1000 * msec








# chan to transfer python string
cdef class StrChan(BaseChan):
    cdef:
        readonly str val, prev_val
        char *cval
        int allocated

    def __init__(self, str name, int max_nelems=1024, **kwargs):
        BaseChan.__init__(self, name, CXDTYPE_TEXT, max_nelems, **kwargs)
        self.cval = <char*>malloc(max_nelems)
        if not self.cval: raise MemoryError()
        self.val,self.prev_val,self.allocated = '', '', 1

    def __dealloc__(self):
        if self.allocated:
            free(self.cval)

    cdef void cb(self):
        self.prev_val = self.val
        c_len = self.get_data(0, self.max_nelems, <void*>self.cval)
        if PY_MAJOR_VERSION > 2:
            self.val = (<bytes>self.cval[:c_len]).decode('UTF-8')
        else:
            self.val = self.cval[:c_len]
        self.valueMeasured.emit(self)
        if self.val != self.prev_val:
            self.valueChanged.emit(self)

    cpdef void setValue(self, str value):
        if PY_MAJOR_VERSION > 2:
            bv = value.encode('UTF-8')
        else:
            bv = unicode(value).encode('UTF-8')

        cdef char *v = bv
        self.snd_data(CXDTYPE_TEXT, len(bv), <void*>v)










from .misc_types cimport *

cdef extern from "cx_common_types.h" nogil:
    ctypedef unsigned long         CxPixel
    #ctypedef struct _CxWidgetType *CxWidget

    ctypedef int CxDataRef_t
    enum:
        NULL_CxDataRef

    ctypedef struct CxKnobParam_t:
        char   *ident
        char   *label
        int     readonly
        int     modified
        double  value
        double  minalwd
        double  maxalwd
        double  rsrvd_d

    ctypedef union CxAnyVal_t:
        float64  f64
        float32  f32
        int64    i64
        uint64   u64
        uint32   u32
        int32    i32
        int16    i16
        uint16   u16
        int8     i8
        uint8    u8
        char8    c8
        char32   c32
